"======================================================================
|
|   Base encodings including Unicode (ISO10646)
|
|
 ======================================================================"


"======================================================================
|
| Copyright 1988-92, 1994-95, 1999, 2000 Free Software Foundation, Inc.
| Written by Paolo Bonzini.
|
| This file is part of the GNU Smalltalk class library.
|
| The GNU Smalltalk class library is free software; you can redistribute it
| and/or modify it under the terms of the GNU Lesser General Public License
| as published by the Free Software Foundation; either version 2.1, or (at
| your option) any later version.
| 
| The GNU Smalltalk class library is distributed in the hope that it will be
| useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
| General Public License for more details.
| 
| You should have received a copy of the GNU Lesser General Public License
| along with the GNU Smalltalk class library; see the file COPYING.LESSER.
| If not, write to the Free Software Foundation, 59 Temple Place - Suite
| 330, Boston, MA 02111-1307, USA.  
|
 ======================================================================"

Smalltalk addSubspace: #Encoders!

Stream subclass: #Encoder
       instanceVariableNames: 'origin'
       classVariableNames: ''
       poolDictionaries: ''
       category: 'i18n-Character sets'!

Encoder subclass: #EncodedStream
       instanceVariableNames: ''
       classVariableNames: 'EncodersRegistry'
       poolDictionaries: 'Encoders'
       category: 'i18n-Character sets'!

Namespace current: Encoders!

Encoder subclass: #FromUCS4
       instanceVariableNames: ''
       classVariableNames: ''
       poolDictionaries: ''
       category: 'i18n-Character sets'!

Encoder subclass: #ToUCS4
       instanceVariableNames: ''
       classVariableNames: ''
       poolDictionaries: ''
       category: 'i18n-Character sets'!

ToUCS4 subclass: #ComposeUCS4LE
       instanceVariableNames: ''
       classVariableNames: ''
       poolDictionaries: ''
       category: 'i18n-Character sets'!

ToUCS4 subclass: #ComposeUCS4BE
       instanceVariableNames: ''
       classVariableNames: ''
       poolDictionaries: ''
       category: 'i18n-Character sets'!

FromUCS4 subclass: #SplitUCS4LE
       instanceVariableNames: 'wch'
       classVariableNames: ''
       poolDictionaries: ''
       category: 'i18n-Character sets'!

FromUCS4 subclass: #SplitUCS4BE
       instanceVariableNames: 'count wch'
       classVariableNames: ''
       poolDictionaries: ''
       category: 'i18n-Character sets'!

Encoder subclass: #Iconv
       instanceVariableNames: 'from to iconvHandle readBuffer readPos readEnd
			       recodedBuffer recodedPos recodedEnd'
       classVariableNames: ''
       poolDictionaries: ''
       category: 'i18n-Character sets'!

FromUCS4 subclass: #ToUTF7
	instanceVariableNames: 'left value lookahead'
	classVariableNames: 'Base64Characters DirectCharacters ToBase64'
	poolDictionaries: ''
	category: 'i18n-Encodings'!

ToUCS4 subclass: #FromUTF7
	instanceVariableNames: 'shift wch lookahead'
	classVariableNames: 'DirectCharacters FromBase64'
	poolDictionaries: ''
	category: 'i18n-Encodings'!

Namespace current: Smalltalk!

!Encoder class methodsFor: 'instance creation'!

on: aStringOrStream from: fromEncoding to: toEncoding
    ^self basicNew
	initializeFrom: fromEncoding to: toEncoding origin: aStringOrStream
! !

!Encoder methodsFor: 'stream operations'!

atEnd
    ^self atEndOfInput
!

next
    ^self nextInput
!

atEndOfInput
    ^origin atEnd
!

nextInput
    ^origin next
! !

!Encoder methodsFor: 'private - initialization'!

initializeFrom: fromEncoding to: toEncoding origin: aStringOrStream
    origin := aStringOrStream isString
	ifTrue: [ aStringOrStream readStream ]
	ifFalse: [ aStringOrStream ]
! !

!EncodedStream class methodsFor: 'initializing'!

initialize
    EncodersRegistry := {
	{ #('UTF7' 'UTF-7'). FromUTF7. ToUTF7 }.
    }
!

registerEncoderFor: arrayOfAliases toUCS4: toUCS4Class fromUCS4: fromUCS4Class
    "Register the two classes that will respectively convert from the
     charsets in arrayOfAliases to UCS4 and vice versa.

     The former class is a stream that accepts characters and returns
     (via #next) integers representing UCS-4 character codes, while
     the latter accepts UCS-4 character codes and converts them to
     characters.  For an example see respectively FromUTF7 and ToUTF7
     (I admit it is not a trivial example)."

    EncodersRegistry := EncodersRegistry copyWith:
	{ arrayOfAliases. toUCS4Class. fromUCS4Class }
! !

!EncodedStream class methodsFor: 'private - triangulating'!

bigEndianPivot
    "When only one of the sides is implemented in Smalltalk
     and the other is obtained via iconv, we use UCS-4 to
     marshal data from Smalltalk to iconv; answer whether we
     should encode UCS-4 characters as big-endian."
    ^Memory bigEndian
!

pivotEncoding
    "When only one of the sides is implemented in Smalltalk
     and the other is obtained via iconv, we need a common
     pivot encoding to marshal data from Smalltalk to iconv.
     Answer the iconv name of this encoding."
    ^self bigEndianPivot ifTrue: [ 'UCS-4BE' ] ifFalse: [ 'UCS-4LE' ]
!

split: input
    "Answer a pipe with the given input stream (which produces
     UCS-4 character codes as integers) and whose output is
     a series of Characters in the required pivot encoding"
    ^self bigEndianPivot
	ifTrue: [ SplitUCS4BE on: input from: 'words' to: 'UCS4-BE' ]
	ifFalse: [ SplitUCS4LE on: input from: 'words' to: 'UCS4-LE' ].
!

compose: input
    "Answer a pipe with the given input stream (which produces
     Characters in the required pivot encoding) and whose output
     is a series of integer UCS-4 character codes."
    ^self bigEndianPivot
	ifTrue: [ ComposeUCS4BE on: input from: 'UCS4-BE' to: 'words' ]
	ifFalse: [ ComposeUCS4LE on: input from: 'UCS4-LE' to: 'words' ].
! !

!EncodedStream class methodsFor: 'instance creation'!

on: aStringOrStream from: fromEncoding
    ^self
	on: aStringOrStream
	from: fromEncoding
	to: Locale default charset
!

on: aStringOrStream to: toEncoding
    ^self
	on: aStringOrStream
	from: aStringOrStream encoding
	to: toEncoding
!

on: aStringOrStream from: fromEncoding to: toEncoding
    | from pivot to encoderFrom encoderTo pipe |

    "Adopt an uniform naming"
    from := fromEncoding asString.
    to := toEncoding asString.
    from = 'UCS-4' ifTrue: [ from := 'UCS-4BE' ].
    to = 'UCS-4' ifTrue: [ to := 'UCS-4BE' ].

    encoderFrom := encoderTo := Iconv.
    EncodersRegistry do: [ :each |
	((each at: 1) includes: to)
	    ifTrue: [ encoderTo := each at: 3 ].
	((each at: 1) includes: from)
	    ifTrue: [ encoderFrom := each at: 2 ]
    ].

    "Let iconv do the triangulation if possible"
    (encoderFrom == Iconv and: [ encoderTo == Iconv ]) ifTrue: [
	^Iconv on: aStringOrStream from: fromEncoding to: toEncoding ].

    "Else answer a `pipe' that takes care of triangulating.
     There is an additional complication: Smalltalk encoders
     read or provide a stream of character codes (respectively
     if the source is UCS-4, or the target is UCS-4), while iconv
     expects raw bytes.  So we add an intermediate layer if
     a mixed Smalltalk+iconv conversion is done: it converts
     character codes --> bytes (SplitUCS4xx, used if iconv will
     convert from UCS-4) or bytes --> character code (ComposeUCS4xx,
     used if iconv will convert to UCS-4).

     There are five different cases (remember that at least one converter
     is not iconv, so `both use iconv' and `from = pivot = to' are banned):
	from = pivot    --> Compose + non-iconv
	pivot = to      --> non-iconv + Split
	to uses iconv   --> non-iconv + Split + iconv (implies from ~= pivot)
	from uses iconv --> iconv + Compose + non-iconv (implies to ~= pivot)
	none uses iconv --> non-iconv + non-iconv (implies neither = pivot)"

    pivot := self pivotEncoding.
    pipe := aStringOrStream.
    from = pivot
	ifFalse: [
	    "Convert to our intermediate representation and split to
	     bytes if needed."

	    pipe := encoderFrom on: pipe from: fromEncoding to: pivot.
	    encoderTo == Iconv ifTrue: [
		pipe := self split: pipe.

		"Check if we already reached the destination format."
		to = pivot ifTrue: [ ^pipe ].
	    ].
	].

    "Compose iconv-produced bytes into UCS-4 character codes if needed."
    encoderFrom == Iconv ifTrue: [
	pipe := self compose: pipe
    ].

    ^encoderTo on: pipe from: pivot to: toEncoding.
! !

Namespace current: Encoders!

!FromUCS4 methodsFor: 'stream operation'!

species
    ^String
! !

!ToUCS4 methodsFor: 'stream operation'!

species
    ^Array
! !

!ComposeUCS4LE methodsFor: 'stream operation'!

next
    ^self nextInput asInteger +
     (self nextInput asInteger bitShift: 8) +
     (self nextInput asInteger bitShift: 16) +
     (self nextInput asInteger bitShift: 24)
! !

!ComposeUCS4BE methodsFor: 'stream operation'!

next
    ^(self nextInput asInteger bitShift: 24) +
     (self nextInput asInteger bitShift: 16) +
     (self nextInput asInteger bitShift: 8) + self nextInput asInteger     
! !

!SplitUCS4LE methodsFor: 'stream operation'!

atEnd
    ^wch == 1 and: [ self atEndOfInput ]
!

next
    | answer |
    wch == 1 ifTrue: [
	"Answer the LSB.  This code will create as few LargeIntegers
         as possible by setting the mark bit only after the LSB has
         been extracted."
	wch := answer := self nextInput.
	wch := (wch bitShift: -8) + 16r1000000.
	^(answer bitAnd: 255) asCharacter
    ].

    "Answer any other byte"
    answer := wch bitAnd: 255.
    wch := wch bitShift: -8.
    ^answer asCharacter
! !

!SplitUCS4LE methodsFor: 'private - initialization'!

initializeFrom: fromEncoding to: toEncoding origin: aStringOrStream
    super initializeFrom: fromEncoding to: toEncoding origin: aStringOrStream.
    wch := 1
! !

!SplitUCS4BE methodsFor: 'stream operation'!

atEnd
    ^count == 0 and: [ self atEndOfInput ]
!

next
    | answer |
    count == 0 ifTrue: [
	"Answer the MSB.  This code will create as few LargeIntegers
         as possible by discarding the bits we answer"
	wch := answer := self nextInput.
	wch := wch bitAnd: 16rFFFFFF.
	count := 3.
	^(answer bitShift: -24) asCharacter
    ].

    "Answer any other byte"
    answer := wch bitShift: -16.
    wch := wch bitAnd: 16rFFFF.
    wch := wch bitShift: 8.
    count := count - 1.
    ^answer asCharacter
! !

!SplitUCS4BE methodsFor: 'private - initialization'!

initializeFrom: fromEncoding to: toEncoding origin: aStringOrStream
    super initializeFrom: fromEncoding to: toEncoding origin: aStringOrStream.
    count := 0.
! !

Iconv
	defineCFunc: 'iconv_open'
	withSelectorArgs: 'iconvOpen: to from: from'
	returning: #cObject
	args: #(#string #string);

	defineCFunc: 'iconv_close'
	withSelectorArgs: 'iconvClose: handle'
	returning: #void
	args: #(#cObject);

	defineCFunc: 'iconvWrapper'
	withSelectorArgs: 'iconvOn: handle from: readBuffer at: readPos
			   size: readCount to: writeBuffer size: writeCount
			   state: bytesLeft'
	returning: #boolean
	args: #(#cObject #smalltalk #int #int #smalltalk #int #smalltalk)!

!Iconv methodsFor: 'stream operation'!

atEnd
    ^self atEndOfBuffer and: [ self convertMore ]
!

next
    | answer |
    self atEndOfBuffer ifTrue: [
	self convertMore ifTrue: [ ^self error: 'end of stream reached' ]
    ].
    answer := recodedBuffer at: recodedPos.
    recodedPos := recodedPos + 1.
    ^answer
!

species
    ^String
! !

!Iconv methodsFor: 'private - living across snapshots'!

update: aspect
    aspect == #aboutToSnapshot ifTrue: [ self release ].
!

release
    self removeToBeFinalized; finalize
!

finalize
    iconvHandle isNil ifTrue: [ ^self ].
    self iconvClose: iconvHandle.
    iconvHandle := nil.
!

iconvOpen
    iconvHandle isNil ifFalse: [ self release ].
    iconvHandle := self iconvOpen: to from: from.
    iconvHandle address = 16rFFFFFFFF
	ifTrue: [ ^self error: 'invalid charset' ].
    self addToBeFinalized.
    ObjectMemory addDependent: self
! !

!Iconv methodsFor: 'private - conversion'!

atEndOfBuffer
    ^recodedPos > recodedEnd
!

refill
    readPos > self refillThreshold ifTrue: [ 
	readBuffer
	    replaceFrom: 1
	    to: readEnd - readPos
	    with: readBuffer
	    startingAt: readPos.

	readEnd := readEnd - readPos.
	readPos := 1
    ].

    [ readEnd >= self bufferSize or: [ self atEndOfInput ] ]
	whileFalse: [
	    readEnd := readEnd + 1.
	    readBuffer at: readEnd put: self nextInput.
	].
!

initializeFrom: fromEncoding to: toEncoding origin: aStringOrStream
    super initializeFrom: fromEncoding to: toEncoding origin: aStringOrStream.
    from := fromEncoding.
    to := toEncoding.
    readPos := 1.
    readEnd := 0.
    recodedPos := 1.
    recodedEnd := 0.
!

bufferSize
    ^1024
!

refillThreshold
    ^1000
!

initBuffers
    readBuffer := String new: self bufferSize.
    recodedBuffer := String new: self bufferSize.
!

convertMore
    | bytesLeft fine |
    recodedBuffer isNil ifTrue: [ self initBuffers ].
    readBuffer isNil ifTrue: [ ^true ].
    iconvHandle isNil ifTrue: [ self iconvOpen ].

    self refill.

    bytesLeft := Array new: 2.
    fine := self
	iconvOn: iconvHandle
	from: readBuffer
	at: readPos
	size: readEnd - readPos + 1
	to: recodedBuffer
	size: self bufferSize
	state: bytesLeft.

    readPos := self bufferSize - (bytesLeft at: 1).
    recodedEnd := self bufferSize - (bytesLeft at: 2).
    recodedPos := 1.

    fine ifFalse: [
	 self error: 'invalid input sequence'.
	 ^false ].

    self atEndOfInput ifFalse: [ ^true ].
    readPos > readEnd
	ifTrue: [ readBuffer := nil ]
	ifFalse: [ self error: 'incomplete sequence at end of stream' ].

    ^false
! !

!ToUTF7 class methodsFor: 'initialization'!

initialize
    Base64Characters := #[
        16r00 16r00 16r00 16r00 16r00 16rA8 16rFF 16r03
        16rFE 16rFF 16rFF 16r07 16rFE 16rFF 16rFF 16r07
    ].

    "Table of direct characters"
    DirectCharacters := #[
        16r00 16r26 16r00 16r00 16r81 16rF3 16rFF 16r87
        16rFE 16rFF 16rFF 16r07 16rFE 16rFF 16rFF 16r07
    ].

    ToBase64 :=
	'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'.
! !

!ToUTF7 methodsFor: 'private'!

isDirect: anInteger
    | d |
    ^anInteger < 128 and: [
	d := DirectCharacters at: (anInteger // 8) + 1.
	((d bitShift: (0 - (anInteger \\ 8))) bitAnd: 1) == 1
    ]
!

toBase64Char: anInteger
    ^ToBase64 at: anInteger + 1
! !

!ToUTF7 methodsFor: 'private'!

atEnd
    ^left == 0 and: [ self atEndOfInput ]
!

flush
    left := 0.
!

next
    | ch |
    left = 0 ifTrue: [
	"Base64 encoding inactive"
	ch := self nextInput.
        (self isDirect: ch) ifTrue: [ ^ch asCharacter ].

	ch = 43
	    ifTrue: [ left := -2. lookahead := $-. ^$+ ].

	ch < 65536
	    ifTrue: [ left := 16. value := ch. ^$+ ].

	ch > 16r110000
	     ifTrue: [ self error: 'invalid input sequence' ].

	"Convert to a surrogate pair"
	ch := ch - 16r10000.
	left := 32.
	value := (((ch bitAnd: 16rFFF) + 16rD800) bitShift: 16) +
		 ((ch bitShift: -10) + 16rDC00).

	^$+
    ].

    left < 0 ifTrue: [
	"left = -2 ---> output the lookahead character and go to 0
	 left = -1 ---> if there's a lookahead output - and go to -2,
			if at end of input output - and stop,
			else reactivate base64"

	left = -2 ifTrue: [
	    left := 0.
	    ch := lookahead asCharacter.
	    lookahead := nil.
	    ^ch
	].
	lookahead isNil ifFalse: [
	    left := -2.
	    ^$-
	].

	left := 0.
	self atEndOfInput ifTrue: [ ^$- ].
    ].

    left < 6 ifTrue: [
	"Pump another character into the Base64 encoder"
        "TODO:  should deactivate base64 encoding"

	(self atEndOfInput or: [
	    self isDirect: (ch := self nextInput) ])
	    ifTrue: [
		"Terminate the stream by left-aligning the last byte"
		value := value bitShift: 6 - (left \\ 6).
		left := left + 6 - (left \\ 6).

		lookahead := ch.
	    ]
	    ifFalse: [
		ch < 65536
		    ifTrue: [
		        left := left + 16.
		        value := (value bitShift: 16) + ch
		    ]
		    ifFalse: [
			ch > 16r110000
			    ifTrue: [ self error: 'invalid input sequence' ].

			"Convert to a surrogate pair"
			ch := ch - 16r10000.
			left := left + 32.
			value := (value bitShift: 32) + 
			     (((ch bitAnd: 16rFFF) + 16rD800) bitShift: 16) +
			     ((ch bitShift: -10) + 16rDC00).
		    ]
	    ]
    ].

    "Take 6 bits out of the Base-64 encoded stream"
    left := left - 6.
    ch := value bitShift: 0 - left.
    value := value bitXor: (ch bitShift: left).

    left = 0 ifTrue: [ left := -1 ].
    ^self toBase64Char: ch
! !

!ToUTF7 methodsFor: 'private - initialization'!

initializeFrom: fromEncoding to: toEncoding origin: aStringOrStream
    super initializeFrom: fromEncoding to: toEncoding origin: aStringOrStream.
    left := 0
! !


!FromUTF7 class methodsFor: 'initialization'!

initialize
    FromBase64 := #[
	62 99 99 99 63
	52 53 54 55 56 57 58 59 60 61 99 99 99 99 99 99 99
	00 01 02 03 04 05 06 07 08 09 10 11 12
	13 14 15 16 17 18 19 20 21 22 23 24 25 99 99 99 99 99 99
	26 27 28 29 30 31 32 33 34 35 36 37 38
	39 40 41 42 43 44 45 46 47 48 49 50 51
     ].

    DirectCharacters := #[
        16r00 16r26 16r00 16r00 16rFF 16rF7 16rFF 16rFF
        16rFF 16rFF 16rFF 16rEF 16rFF 16rFF 16rFF 16r3F
        16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00
        16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00
    ].
!

!FromUTF7 methodsFor: 'private'!

isDirect: anInteger
    | d |
    ^anInteger < 128 and: [
	d := DirectCharacters at: anInteger // 8 + 1.
	((d bitShift: (0 - (anInteger \\ 8))) bitAnd: 1) == 1
    ]
!

fromBase64Char: aCharacter
    ^FromBase64 at: aCharacter - 42
! !

!FromUTF7 methodsFor: 'converting'!

atEnd
    ^lookahead isNil
!

next
    | ch |
    "The decoder will always decode a character ahead, because when we
     are to read only a minus, we might already be at the end of the
     stream! Here is a simple example: +AFs- which decodes to [
	 We read + and switch to base-64 --> shift = 32
	 We read A and put it into the accumulator --> shift = 26
	 We read F and put it into the accumulator --> shift = 20
	 We read s and put it into the accumulator --> shift = 14

     We then decode the [ and return it.  Now we are not
     #atEndOfInput yet, but there are no more characters to
     give away!  Since we are not sure that the source supports
     #peek, our only other option would be to implement peeking
     for it and check for $- now.  This would have an overhead
     proportional to the number of input characters (to check
     whether we have already peeked the next characters), while
     our choice's overhead is proportional to the number of output
     characters, which is always less in UTF-7."

    lookahead isNil ifTrue: [ self error: 'end of data reached' ].
    ch := lookahead.
    lookahead := self getNext.
    ^ch
!

getNext
    | ch value wc1 |
    self atEndOfInput ifTrue: [ ^nil ].
    ch := self nextInput value.
    shift = 0 ifTrue: [
        (self isDirect: ch) ifTrue: [ ^ch ].
	ch == 43 "plus" ifFalse: [ self error: 'invalid input sequence' ].

	ch := self nextInput value.
	ch == 45 "minus" ifTrue: [ ^43 ].

	"Else switch into base64 mode"
	shift := 32.
	wch := 0.
    ].

    ((ch between: 43 and: 122) and: [
	(value := self fromBase64Char: ch) < 99 ])
	ifFalse: [
	    "Terminate base64 encoding.
	     If accumulated data is nonzero, the input is invalid.
	     Also, partial UTF-16 characters are invalid."
	    (shift <= 26 or: [ wch > 0 ]) ifTrue: [
		shift := 0.
		self error: 'invalid input sequence' ].

	    shift := 0.
	    ch = 45 "minus" ifFalse: [ ^ch ].

	    "Discard the -"
	    ^self getNext
	].

    "Concatenate the base64 integer value to the accumulator"
    shift > 6 ifTrue: [
	shift := shift - 6.
	wch := wch + (value bitShift: shift).
	shift > 16 ifTrue: [ ^self getNext ].

	(shift between: 11 and: 16) ifTrue: [
	    "Completed an UTF-16 character.  When we see a High
	     Surrogate, we must wait for the following Low Surrogate."
	    wc1 := wch bitShift: -16.

	    (wc1 between: 16rD800 and: 16rDBFF) ifFalse: [
		wch := (wch bitAnd: 65535) bitShift: 16.
		shift := shift + 16.
		^wc1
	    ].
	].

	(shift between: 5 and: 10) ifTrue: [
	    "After an High Surrogate, verify that the next character
	     is indeed a Low Surrogate"
	    wc1 := wch bitAnd: 65535.
	    (wc1 between: 16rDC00 and: 16rDFFF) ifFalse: [
		shift := 0.
		self error: 'invalid input sequence' ]
	].
    ].

    "Completed an UTF-16 surrogate pair"

    "35FDC00 = -0xD800 << 10 - 0xDC00 + 0x10000"
    wc1 := wch bitShift: -16.
    wch := (wch bitAnd: 65535) + (value bitShift: shift - 6).
    wc1 := (wc1 bitShift: 10) + wch - 16r35FDC00.

    wch := ((value bitShift: shift) bitAnd: 16r3F) bitShift: 26.
    ^wc1
! !

!FromUTF7 methodsFor: 'private - initialization'!

initializeFrom: fromEncoding to: toEncoding origin: aStringOrStream
    super initializeFrom: fromEncoding to: toEncoding origin: aStringOrStream.
    shift := 0.
    lookahead := self getNext.
! !


Namespace current: Smalltalk!

"Now add some extensions to the system classes"

(CharacterArray classPool includesKey: #DefaultEncoding)
    ifFalse: [ CharacterArray addClassVarName: #DefaultEncoding ]!

!CharacterArray class methodsFor: 'converting'!

defaultEncoding
    DefaultEncoding isNil
	ifTrue: [ DefaultEncoding := Locale default charset ].

    ^DefaultEncoding
!

defaultEncoding: aString
    DefaultEncoding := aString
! !

!CharacterArray methodsFor: 'converting'!

encoding
    ^self class defaultEncoding asString = 'UTF-16'
	ifTrue: [ self utf8Encoding ]
	ifFalse: [ self class defaultEncoding ]
!

utf16Encoding
    | b1 b2 bigEndian |
    b1 := self at: 1.		"Low byte"
    b2 := self at: 2.		"High byte"

    "If we read FFFE, it is big-endian; if we read FEFF, it is little-endian"
    bigEndian := b1 * b2 = 16rFD02
	ifTrue: [ b1 = 255 ]
	ifFalse: [ Memory bigEndian ].

    ^bigEndian ifTrue: [ 'UTF-16BE' ] ifFalse: [ 'UTF-16LE' ]
! !

!ReadStream methodsFor: 'converting'!

encoding
    ^collection encoding
! !

Encoders ToUTF7 initialize!
Encoders FromUTF7 initialize!
EncodedStream initialize!
